import argparse
import ast
import inspect
import logging
import os
import sys

import app.modules as modules
from app.helpers import get_module_variables, import_module, walk


def main():
    ap = argparse.ArgumentParser(description='')

    ap.add_argument('-p', '--path-to-schemas', nargs='?', type=str,
                    help='name of folder in current directory containing schemas, default: schemas',
                    default='schemas')
    ap.add_argument('-v', '--verbose',
                    help='increase output verbosity',
                    action='store_true')
    ap.add_argument('-a', '--all',
                    help='generate overloads for all standard schema types',
                    action='store_true')

    args = ap.parse_args()
    log_level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(level=log_level)

    is_add_all = args.all

    cwd = os.getcwd()
    sys.path.append(cwd)

    file_count = 0
    schemas_count = 0
    schemas_errors_count = 0

    blahblah_module = modules.BlahBlahModule()

    if is_add_all:
        logging.debug('.. creating standard types overload')
        blahblah_module.generate_standard_types()

    for file_name in walk(args.path_to_schemas):
        logging.debug(f'.. creating types for: {file_name}')
        module = import_module(file_name)
        module_source = inspect.getsource(module)
        module_ast = ast.parse(module_source)

        typed_module = modules.TypedModule(file_name)

        for name in get_module_variables(module_ast):
            value = getattr(module, name)

            try:
                typed_module.generate(name, value)
                blahblah_module.generate(file_name, name, value)
                schemas_count += 1
            except Exception:
                logging.error(f'Failed typing schema {name}, skipping')
                schemas_errors_count += 1

        typed_module.print()
        file_count += 1

    blahblah_module.print()

    logging.info(
        f'Successfully processed {schemas_count} schemas, {file_count} files in {args.path_to_schemas}/')

    if schemas_errors_count:
        logging.info(
            f'Failed processed {schemas_errors_count} schemas')

    sys.path.remove(cwd)


if __name__ == '__main__':
    main()



import app.modules as modules
from app.helpers import load_module_from_string

"""
Не покрыт:
 - unordered
 - uuid
"""

CODE = '''
from district42_exp_types.numeric import schema_numeric
TestNumericSchema =  schema_numeric.min(1)
'''

CODE_PYI = '''\
from district42_exp_types.numeric import NumericSchema
TestNumericSchema: NumericSchema\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from district42_exp_types.numeric import NumericSchema

@overload
def fake(schema: NumericSchema) -> str:
    pass\
'''


def test_numeric_pyi():
    module = load_module_from_string('test', CODE)

    schema_name = 'TestNumericSchema'
    schema_description = getattr(module, schema_name)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(schema_name, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_numeric_pyi_blahblah():
    module = load_module_from_string('test', CODE)

    schema_name = 'TestNumericSchema'
    schema_description = getattr(module, schema_name)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test_file_name', schema_name, schema_description)

    assert blahblah_module.get_printable_content() == CODE_BLAHBLAH_PYI



from .files import walk
from .files_as_modules import get_module_variables, import_module, load_module_from_string
from .schemas import (
    get_module_to_import_from,
    get_types_from_any,
    is_builtin_class_instance,
    is_schema_type_simple,
)
from .schemas_dict import (
    has_invalid_key,
    is_dict_empty,
    is_dict_typed_as_empty,
    is_dict_without_keys,
)



class Typing:

    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.class_name = self.value.__class__.__name__
        self.class_type = self.value.__class__.type

    def generate_pyi(self): ...

    def generate_fake_overload(self,  path_to_schema: str): ...


class UnknownTypeSchema:

    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.class_name = self.value.__class__.__name__


class OverloadedFake:
    def __init__(self, schema_type, ast_method):
        # schema_type - тип описываемой схемы = входного параметра
        self.schema_type = schema_type
        self.ast_method = ast_method

    def __eq__(self, other):
        return self.schema_type == other.schema_type



from niltype import Nil


def has_invalid_key(dictionary: dict) -> bool:
    for key in dictionary.keys():
        if not key.isidentifier():
            return True
    return False


def is_dict_without_keys(dict_value) -> bool:
    return dict_value.props.keys is Nil


def is_dict_empty(dict_value) -> bool:
    return dict_value.props.keys == {}


def is_dict_typed_as_empty(dict_value) -> bool:
    return (
            is_dict_without_keys(dict_value)
            or is_dict_empty(dict_value)
            or has_invalid_key(dict_value.props.keys)
    )



from d42 import schema
from d42.custom_type import Schema

from app.modules.module import Module
from app.types import (
    AnyTyping,
    CustomTyping,
    DictTyping,
    ListTyping,
    OverloadedFake,
    ScalarTyping,
    SchemaTyping,
    UnknownTypeSchema,
)


class BlahBlahModule(Module):
    def __init__(self):
        super().__init__('blahblah.py')
        self.add_import('typing', 'overload')
        self.overloaded_fakes = []

    def _add_overload(self, method: OverloadedFake):
        if method is not None and method not in self.overloaded_fakes:
            self.overloaded_fakes.append(method)

    def get_ast_content(self) -> list | None:
        imports = [import_.to_ast() for import_ in self.imports]
        methods = [fake.ast_method for fake in self.overloaded_fakes]
        if methods:
            return imports + methods
        return None

    def generate_standard_types(self):
        standard_schemas = [
            (schema.list, ListTyping),
            (schema.bool, ScalarTyping),
            (schema.str, ScalarTyping),
            (schema.int, ScalarTyping),
            (schema.any, AnyTyping),
            (schema.dict, DictTyping),
            (schema.float, ScalarTyping),
            (schema.none, ScalarTyping),
        ]

        for schema_, class_typing in standard_schemas:
            overload, imports = class_typing(schema_.__class__.__name__,
                                             schema_).generate_fake_overload('')
            self.add_import_new(*imports)
            self._add_overload(overload)

        overload, imports = SchemaTyping(Schema).generate_fake_overload()
        self.add_import_new(*imports)
        self._add_overload(overload)

    def generate(self, file_name, schema_name, schema_value):
        if not isinstance(schema_value, Schema):
            return

        if schema_name.startswith('_'):
            return

        schema_for_typing = UnknownTypeSchema(schema_name, schema_value)

        schema_typings = [
            CustomTyping,
            ScalarTyping,
            DictTyping,
            AnyTyping,
            ListTyping,
        ]

        for typing_ in schema_typings:
            if typing_.is_valid_type_for_schema(schema_for_typing):
                overload, imports = typing_(schema_name, schema_value).generate_fake_overload(file_name)
                self.add_import_new(*imports)
                self._add_overload(overload)



from .annotated_assign import annotated_assign, annotated_assign_union
from .dict_metaclass import dict_metaclass
from .dict_typeclass import dict_typeclass
from .fake_overload import (
    fake_dict_overload,
    fake_list_overload,
    fake_none_overload,
    fake_scalar_overload,
    fake_union_overload,
)
from .list_typeclass import list_typeclass



from ._type import OverloadedFake, UnknownTypeSchema
from .any import AnyTyping
from .custom import CustomTyping
from .dict import DictTyping
from .list import ListTyping
from .scalar import ScalarTyping
from .schema import SchemaTyping



import sys

import app.modules as modules
from app.helpers import load_module_from_string
from tests.custom_schema_with_type._text_schema import TEXT_SCHEMA_CODE

SCHEMA_NAME = 'TestSchema'

CODE = '''\
from custom import TextSchema
from d42 import schema

TestSchema = TextSchema | schema.str
'''

CODE_PYI = '''\
from district42.types import StrSchema
TestSchema: StrSchema\
'''

BLAHBLAH_PYI = '''\
from typing import overload
from district42.types import StrSchema

@overload
def fake(schema: StrSchema) -> str:
    pass\
'''


class TestClassTextSchema:

    def setup_method(self):
        entity_module = load_module_from_string('test_entity', TEXT_SCHEMA_CODE)
        sys.modules['custom'] = entity_module

    def teardown_method(self):
        sys.modules.pop('custom')

    def test_any_schema_custom_pyi(self):
        module = load_module_from_string('test_scalar', CODE)
        schema_value = getattr(module, SCHEMA_NAME)

        typed_module = modules.TypedModule('file_name')
        typed_module.generate(SCHEMA_NAME, schema_value)

        assert typed_module.get_printable_content() == CODE_PYI

    def test_any_schema_custom_pyi_blahblah(self):
        module = load_module_from_string('test.module', CODE)
        schema_value = getattr(module, SCHEMA_NAME)

        blahblha_module = modules.BlahBlahModule()
        blahblha_module.generate('test.module', SCHEMA_NAME, schema_value)

        assert blahblha_module.get_printable_content() == BLAHBLAH_PYI



import os
from typing import Generator


def walk(directory: str) -> Generator[str, None, None]:
    for root, _, files in os.walk(directory):
        for file in sorted(files):
            if not file.endswith(".py") or file == "__init__.py":
                continue
            yield os.path.join(root, file)



import app.modules as modules
from app.helpers import load_module_from_string

CODE = '''\
from d42 import schema
TestSchema = schema.dict({
    '100x100': schema.str
})\
'''

CODE_PYI = '''\
from district42.types import DictSchema
TestSchema: DictSchema\
'''

CODE_BLAHBLAH = '''\
from typing import overload
from typing import Dict
from district42.types import DictSchema

@overload
def fake(schema: DictSchema) -> Dict:
    pass\
'''


def test_dict_key_not_identifier_pyi():
    module = load_module_from_string('test_scalar', CODE)

    schema_name = 'TestSchema'
    schema_description = getattr(module, schema_name)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(schema_name, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_dict_key_not_identifier_blahblah_pyi():
    module = load_module_from_string('test.module', CODE)

    schema_name = 'TestSchema'
    schema_description = getattr(module, schema_name)

    blahblha_module = modules.BlahBlahModule()
    blahblha_module.generate('test.module', schema_name, schema_description)

    assert blahblha_module.get_printable_content() == CODE_BLAHBLAH



import typing
from typing import Tuple

import ast_generate
from app.helpers import get_module_to_import_from
from app.modules.module import Import
from app.types._type import OverloadedFake


class SchemaTyping:

    def __init__(self, value):
        self.value = value
        self.class_name = self.value.__name__

    def generate_fake_overload(self) -> Tuple[OverloadedFake, list[Import]]:
        imports = []

        imports.append(Import(get_module_to_import_from(self.value), self.class_name))
        imports.append(Import('typing', 'Any'))
        overload = OverloadedFake(
            typing.Any,
            ast_generate.fake_scalar_overload(self.class_name, typing.Any)
        )
        return overload, imports



import ast


def annotated_assign(name: str, type_: str) -> ast.AnnAssign:
    """return annotation assign like: name: type"""
    return ast.AnnAssign(
        target=ast.Name(id=name),
        annotation=ast.Name(id=type_),
        simple=1,
    )


def annotated_assign_union(schema_name: str, types_in_union: list[str | None]) -> ast.AnnAssign:
    return ast.AnnAssign(
        target=ast.Name(id=schema_name),
        annotation=ast.Subscript(
            value=ast.Name(id='Union'),
            slice=ast.Tuple(
                elts=[
                    ast.Name(id=type_.__name__) for type_ in types_in_union
                ],
            ),
        ),
        simple=1,
    )



import ast
import importlib.util
from types import ModuleType


def import_module(file_path: str) -> ModuleType:
    name = file_path[:-3].replace("/", ".")
    spec = importlib.util.spec_from_file_location(name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_module_variables(module_ast: ast.Module) -> list[str]:
    variables = []

    def process_target(target: ast.expr):
        assert isinstance(target, ast.Name)
        if target.id.startswith("__"):
            return
        variables.append(target.id)

    for node in module_ast.body:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                process_target(target)
        elif isinstance(node, ast.AnnAssign):
            process_target(node.target)
    return variables


def load_module_from_string(module_name, module_content):
    spec = importlib.util.spec_from_loader(module_name, loader=None, origin='string')

    module = importlib.util.module_from_spec(spec)

    exec(module_content, module.__dict__)
    return module



import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'NestedTestSchema'

CODE = '''\
from d42 import schema

NestedTestSchema = schema.dict({
    'data': schema.dict({})
})
'''

CODE_PYI = '''\
from typing import overload
from typing import Literal
from typing import TypedDict
from typing import Dict

class _D42MetaNestedTestSchema(type):

    def __getitem__(cls, arg: Literal['data']) -> Dict:
        pass

    def __mod__(self, other):
        pass

    def __add__(self, other):
        pass

class NestedTestSchema(metaclass=_D42MetaNestedTestSchema):

    class type(TypedDict, total=False):
        data: Dict\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from typing import Type
from test.module import NestedTestSchema

@overload
def fake(schema: Type[NestedTestSchema]) -> NestedTestSchema.type:
    pass\
'''


def test_dict_nested_empty_dict_key_pyi():
    module = load_module_from_string('test_scalar', CODE)

    schema_value = getattr(module, SCHEMA_NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(SCHEMA_NAME, schema_value)

    assert typed_module.get_printable_content() == CODE_PYI


def test_dict_nested_empty_dict_key_pyi_blahblah():
    module = load_module_from_string('test.module', CODE)

    schema_value = getattr(module, SCHEMA_NAME)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test.module', SCHEMA_NAME, schema_value)

    assert blahblah_module.get_printable_content() == CODE_BLAHBLAH_PYI



import ast
import typing
from typing import Tuple

from d42.custom_type import CustomSchema
from district42 import Schema

import ast_generate
from app.helpers import get_module_to_import_from
from app.modules.module import Import
from app.types._type import OverloadedFake, Typing, UnknownTypeSchema
from ast_generate import annotated_assign


class CustomTyping(Typing):

    def __init__(self, name, value):
        super().__init__(name, value)
        self.class_name = self.value.type.__name__

    @classmethod
    def is_valid_type_for_schema(cls, schema_: UnknownTypeSchema) -> bool:
        return issubclass(schema_.value.__class__, CustomSchema)

    def generate_pyi(self) -> Tuple[list[ast.AnnAssign], list[Import]]:
        # для схем без типов
        if (
                hasattr(self.value, 'type') is False
                or self.value.type is typing.Any
        ):
            return [], []
        annotation = annotated_assign(self.name, self.class_name)
        imports = [Import(get_module_to_import_from(self.value.type), self.class_name)]
        return [annotation], imports

    def generate_fake_overload(self,  path_to_schema: str) -> Tuple[OverloadedFake, list[Import]]:
        # для кастомных схем, у которых не прописан тип:
        if (
            hasattr(self.value, 'type') is False
            or self.value.type is typing.Any
        ):
            registered_type_name = self.value.__class__.__name__
            imports = [
                Import('typing', 'Any'),
                Import(get_module_to_import_from(self.value), registered_type_name)
            ]
            overload = OverloadedFake(typing.Any,
                                      ast_generate.fake_scalar_overload(
                                          registered_type_name, typing.Any))
            return overload, imports

        else:
            if issubclass(self.class_type, Schema):
                imports = [
                    Import(get_module_to_import_from(self.class_type), self.class_name)
                ]
                overload = OverloadedFake(self.class_name,
                                          ast_generate.fake_scalar_overload(self.class_name,
                                                                            self.class_type.type))
                return overload, imports
            else:
                # todo
                pass
            pass
        return None, []



import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'TestSchema'

CODE = '''\
from d42 import schema

TestSchema = schema.dict({
    "id": schema.any,
})
'''

CODE_PYI = '''\
from typing import overload
from typing import Literal
from typing import TypedDict
from district42.types import AnySchema

class _D42MetaTestSchema(type):

    def __getitem__(cls, arg: Literal['id']) -> AnySchema:
        pass

    def __mod__(self, other):
        pass

    def __add__(self, other):
        pass

class TestSchema(metaclass=_D42MetaTestSchema):

    class type(TypedDict, total=False):
        id: AnySchema.type\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from typing import Type
from test.module import TestSchema

@overload
def fake(schema: Type[TestSchema]) -> TestSchema.type:
    pass\
'''


def test_dict_key_any_schema_pyi():
    module = load_module_from_string('test', CODE)
    schema_value = getattr(module, SCHEMA_NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(SCHEMA_NAME, schema_value)

    assert typed_module.get_printable_content() == CODE_PYI


def test_dict_key_key_any_schema_pyi_blahblah():
    module = load_module_from_string('test.module', CODE)
    schema_value = getattr(module, SCHEMA_NAME)

    blahblha_module = modules.BlahBlahModule()
    blahblha_module.generate('test.module', SCHEMA_NAME, schema_value)

    assert blahblha_module.get_printable_content() == CODE_BLAHBLAH_PYI



import typing

from d42.custom_type import Schema

from app.helpers import is_builtin_class_instance
from app.types import (
    AnyTyping,
    CustomTyping,
    DictTyping,
    ListTyping,
    ScalarTyping,
    UnknownTypeSchema,
)
from ast_generate import annotated_assign

from .module import Module


class TypedModule(Module):

    def _add_typing(self, item):
        self.typed_items.append(item)

    def _add_typings(self, items):
        for item in items:
            self.typed_items.append(item)

    def get_ast_content(self) -> list[str] | None:
        imports = [import_.to_ast() for import_ in self.imports]
        items = self.typed_items
        if items:
            return imports + items
        return None

    def generate(self, schema_name: str, schema_value: typing.Any):
        if not isinstance(schema_value, Schema):
            if is_builtin_class_instance(type(schema_value)):
                self._add_typing(annotated_assign(schema_name, type(schema_value).__name__))
            return

        schema_for_typing = UnknownTypeSchema(schema_name, schema_value)

        schema_typings = [
            CustomTyping,
            ScalarTyping,
            DictTyping,
            AnyTyping,
            ListTyping,
        ]

        for typing_ in schema_typings:
            if typing_.is_valid_type_for_schema(schema_for_typing):
                annotations, imports = typing_(schema_name, schema_value).generate_pyi()
                self.add_import_new(*imports)
                self._add_typings(annotations)
                return



import app.modules as modules
from app.helpers import load_module_from_string

CODE = '''\
from district42_exp_types.uuid_str import schema_uuid_str
StrUuidExampleSchema = schema_uuid_str
'''

CODE_PYI = '''\
from district42_exp_types.uuid_str import UUIDStrSchema
StrUuidExampleSchema: UUIDStrSchema\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from district42_exp_types.uuid_str import UUIDStrSchema
from typing import Any

@overload
def fake(schema: UUIDStrSchema) -> Any:
    pass\
'''


def test_str_uuid_pyi():
    module = load_module_from_string('test', CODE)

    schema_name = 'StrUuidExampleSchema'
    schema_description = getattr(module, schema_name)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(schema_name, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_str_uuid_pyi_blahblah():
    module = load_module_from_string('test', CODE)

    schema_name = 'StrUuidExampleSchema'
    schema_description = getattr(module, schema_name)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test_file_name', schema_name, schema_description)

    assert blahblah_module.get_printable_content() == CODE_BLAHBLAH_PYI



import app.modules as modules
from app.helpers import load_module_from_string

CODE = '''\
from d42 import schema
from jj_district42 import HistoryItemSchema, schema_history_request

TestHistorySchema = HistoryItemSchema + schema.dict({
    "request": schema_history_request
})
'''

CODE_PYI = '''\
from typing import overload
from typing import Literal
from typing import TypedDict
from typing import Any
from typing import List
from district42.types import DateTimeSchema

class _D42MetaTestHistorySchema(type):

    @overload
    def __getitem__(cls, arg: Literal['request']) -> Any:
        pass

    @overload
    def __getitem__(cls, arg: Literal['response']) -> Any:
        pass

    @overload
    def __getitem__(cls, arg: Literal['tags']) -> List:
        pass

    @overload
    def __getitem__(cls, arg: Literal['created_at']) -> DateTimeSchema:
        pass

    def __mod__(self, other):
        pass

    def __add__(self, other):
        pass

class TestHistorySchema(metaclass=_D42MetaTestHistorySchema):

    class type(TypedDict, total=False):
        request: Any
        response: Any
        tags: List
        created_at: DateTimeSchema.type\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from typing import Type
from test_file_name import TestHistorySchema

@overload
def fake(schema: Type[TestHistorySchema]) -> TestHistorySchema.type:
    pass\
'''


def test_jj_history_pyi():
    module = load_module_from_string('test', CODE)

    schema_name = 'TestHistorySchema'
    schema_description = getattr(module, schema_name)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(schema_name, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_jj_history_pyi_blahblah():
    module = load_module_from_string('test', CODE)

    schema_name = 'TestHistorySchema'
    schema_description = getattr(module, schema_name)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test_file_name', schema_name, schema_description)

    assert blahblah_module.get_printable_content() == CODE_BLAHBLAH_PYI



import ast
import logging


class Import:
    def __init__(self, module, item):
        self.module = module
        self.item = item

    def __eq__(self, other):
        return self.module == other.module and self.item == other.item

    def __repr__(self):
        return f'from {self.module} import {self.item}'

    def to_ast(self):
        return ast.ImportFrom(module=self.module, names=[ast.alias(name=self.item, asname=None)], level=0)


class Module:
    imports: list

    def __init__(self, path: str):
        self.path = path
        self.imports = []
        self.typed_items = []

    def no_import_duplicates(self, import_: Import) -> bool:
        return import_ not in self.imports

    def duplicate_in_module(self, import_: Import):
        if import_.item in [i.item for i in self.imports]:
            return import_.module
        return None

    def add_import(self, module, *imported_items):
        # todo squash imports
        for item in imported_items:
            new_import = Import(module, item)
            if self.no_import_duplicates(new_import):

                duplicate_module = self.duplicate_in_module(new_import)
                if duplicate_module:
                    logging.warning(f'Item {new_import.item} has already imported from module {duplicate_module}')

                self.imports.append(new_import)

    def add_import_new(self, *imports):
        for import_ in imports:
            if self.no_import_duplicates(import_):

                duplicate_module = self.duplicate_in_module(import_)
                if duplicate_module:
                    logging.warning(f'Item {import_.item} has already imported from module {duplicate_module}')

                self.imports.append(import_)

    def get_ast_content(self) -> list:
        return []

    def get_printable_content(self) -> str | None:
        module = ast.Module(body=self.get_ast_content(), type_ignores=[])
        if module.body is None:
            return None
        return ast.unparse(ast.fix_missing_locations(module))  # Python 3.9+

    def print(self):
        content = self.get_printable_content()
        if content is not None:
            with open(f"{self.path}i", "w") as f:
                f.write(content)



import app.modules as modules

BLAHBLAH_PYI_STANDARD_TYPES = """\
from typing import overload
from typing import List
from district42.types import ListSchema
from district42.types import BoolSchema
from district42.types import StrSchema
from district42.types import IntSchema
from typing import Any
from district42.types import AnySchema
from typing import Dict
from district42.types import DictSchema
from district42.types import FloatSchema
from district42.types import NoneSchema
from district42.types import Schema

@overload
def fake(schema: ListSchema) -> List:
    pass

@overload
def fake(schema: BoolSchema) -> bool:
    pass

@overload
def fake(schema: StrSchema) -> str:
    pass

@overload
def fake(schema: IntSchema) -> int:
    pass

@overload
def fake(schema: AnySchema) -> Any:
    pass

@overload
def fake(schema: DictSchema) -> Dict:
    pass

@overload
def fake(schema: FloatSchema) -> float:
    pass

@overload
def fake(schema: NoneSchema) -> None:
    pass

@overload
def fake(schema: Schema) -> Any:
    pass\
"""


def test_all_mode_pyi_blahblah():

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate_standard_types()

    # todo проверять иначе
    assert blahblah_module.get_printable_content() == BLAHBLAH_PYI_STANDARD_TYPES



import ast


def list_typeclass(name: str, item_type: str):
    """
    class NameSchema: type = List[item_type.type]
    """
    return ast.ClassDef(
        name=f"{name}",
        bases=[],
        body=[
            ast.Assign(
                targets=[ast.Name(id='type')],
                value=ast.Subscript(
                    value=ast.Name(id='List'),
                    slice=ast.Attribute(
                        value=ast.Name(id=item_type),
                        attr='type',
                    )
                ),
            )
        ],
        keywords=[],
        decorator_list=[],
        starargs=None,
        kwargs=None,
    )



import ast
import typing
from typing import Tuple

import ast_generate
from app.helpers import get_module_to_import_from
from app.modules.module import Import
from app.types._type import OverloadedFake, Typing, UnknownTypeSchema
from ast_generate import annotated_assign


class ScalarTyping(Typing):

    @classmethod
    def is_valid_type_for_schema(cls, schema_: UnknownTypeSchema) -> bool:
        scalar_types = [
            'StrSchema', 'BoolSchema', 'IntSchema', 'FloatSchema', 'NoneSchema',
            'BytesSchema', 'NoneSchema', 'DateTimeSchema',
            'NumericSchema', 'UUIDStrSchema', 'UUID4Schema'
        ]
        return schema_.class_name in scalar_types

    def generate_pyi(self) -> Tuple[list[ast.AnnAssign], list[Import]]:
        annotation = annotated_assign(self.name, type(self.value).__name__)
        imports = [Import(get_module_to_import_from(self.value), self.class_name)]
        return [annotation], imports

    def generate_fake_overload(self,  path_to_schema: str) -> Tuple[OverloadedFake, list[Import]]:
        imports = []

        imports.append(Import(get_module_to_import_from(self.value), self.class_name))
        if self.class_name == 'NoneSchema':
            overload = OverloadedFake(
                self.class_name,
                ast_generate.fake_none_overload(self.class_name)
            )
            return overload, imports

        elif self.value.type is typing.Any:
            imports.append(Import('typing', 'Any'))
            overload = OverloadedFake(
                typing.Any,
                ast_generate.fake_scalar_overload(self.class_name, typing.Any)
            )
            return overload, imports

        overload = OverloadedFake(
            self.class_name,
            ast_generate.fake_scalar_overload(self.class_name, self.value.__class__.type)
        )

        return overload, imports



import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'TestSchema'

CODE = '''\
from d42 import schema
TestSchema = [schema.str("string")]
'''


CODE_PYI = '''\
TestSchema: list\
'''


def test_list_of_schemas_pyi():
    module = load_module_from_string('test', CODE)
    schema_description = getattr(module, SCHEMA_NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(SCHEMA_NAME, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_list_of_schemas_pyi_blahblah():
    module = load_module_from_string('test', CODE)
    schema_description = getattr(module, SCHEMA_NAME)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test_file_name', SCHEMA_NAME, schema_description)

    assert blahblah_module.get_printable_content() is None



import ast
from typing import Tuple

from niltype import Nil

import ast_generate
from app.helpers import get_module_to_import_from, get_types_from_any
from app.modules.module import Import
from app.types._type import OverloadedFake, Typing, UnknownTypeSchema


class AnyTyping(Typing):

    @classmethod
    def is_valid_type_for_schema(cls, schema_: UnknownTypeSchema) -> bool:
        return schema_.class_name == 'AnySchema'

    def generate_pyi(self) -> Tuple[list[ast.AnnAssign], list[Import]]:
        imports = []
        annotations = []

        if hasattr(self.value.props, 'types') and self.value.props.types is not Nil:
            types_in_any = get_types_from_any(self.value.props)

            if len(types_in_any) == 1:
                type_ = types_in_any.pop()

                if type_.__name__ == 'DictSchema':
                    ...
                    # todo union для словарей
                imports.append(Import(get_module_to_import_from(type_), type_.__name__))
                annotations.append(ast_generate.annotated_assign(self.name, type_.__name__))
                return annotations, imports

            else:
                imports.append(Import('typing', 'Union'))
                for type_ in types_in_any:
                    imports.append(Import(get_module_to_import_from(type_), type_.__name__))
                annotations.append(ast_generate.annotated_assign_union(self.name, types_in_any))
                return annotations, imports

        imports.append(Import('district42.types', 'AnySchema'))
        annotations.append(ast_generate.annotated_assign(self.name, 'AnySchema'))
        return annotations, imports

    def generate_fake_overload(self,  path_to_schema: str) -> Tuple[OverloadedFake, list[Import]]:
        imports = []
        if self.value.props.types is not Nil:
            types_in_any = get_types_from_any(self.value.props)

            if len(types_in_any) == 1:
                type_ = types_in_any[0]
                class_name = type_.__name__
                imports.append(Import(get_module_to_import_from(type_), class_name))
                overload = OverloadedFake(
                    class_name,
                    ast_generate.fake_scalar_overload(class_name, type_.type)
                )
                return overload, imports

            else:
                module_name = path_to_schema.replace('/', '.').replace('.py', '')
                imports.extend([
                    Import('typing', 'Union'),
                    Import(module_name, self.name),
                ])
                overload = OverloadedFake(
                    self.name,
                    ast_generate.fake_union_overload(self.name, types_in_any)
                )
                return overload, imports

        imports.append(Import('typing', 'Any'))
        imports.append(Import(get_module_to_import_from(self.value), 'AnySchema'))
        overload = OverloadedFake(
            'AnySchema',
            ast_generate.fake_scalar_overload('AnySchema', self.value.__class__.type)
        )
        return overload, imports



import app.modules as modules
from app.helpers import load_module_from_string

NAME = 'TestUUID4Schema'

CODE = '''\
from d42 import schema
TestUUID4Schema = schema.uuid4
'''

CODE_PYI = '''\
from district42.types import UUID4Schema
TestUUID4Schema: UUID4Schema\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from district42.types import UUID4Schema
from typing import Any

@overload
def fake(schema: UUID4Schema) -> Any:
    pass\
'''


def test_scalar_uuid4_pyi():
    module = load_module_from_string('test', CODE)
    schema_description = getattr(module, NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(NAME, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_scalar_uuid4_pyi_blahblah():
    module = load_module_from_string('test', CODE)
    schema_description = getattr(module, NAME)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test_file_name', NAME, schema_description)

    assert blahblah_module.get_printable_content() == CODE_BLAHBLAH_PYI



import ast
from typing import Any

from ast_generate.dict_value_type import AnyDictValueType


def generate_key_typing_assign(name: str, type_: Any) -> ast.AnnAssign:
    return ast.AnnAssign(
        target=ast.Name(id=name),
        annotation=type_.get_type_for_typeclass(),
        value=None,
        simple=1
    )


def dict_typeclass(name: str, meta_name: str, typing_map: dict[str, AnyDictValueType]):
    return ast.ClassDef(
        name=f"{name}",
        bases=[],
        body=[
            ast.ClassDef(
                name="type",
                bases=[ast.Name(id="TypedDict", ctx=ast.Load())],
                keywords=[ast.keyword(arg='total', value=ast.NameConstant(value=False))],
                body=[
                    generate_key_typing_assign(name, type_)
                    for name, type_ in typing_map.items()
                ],
                decorator_list=[],
            )
        ],
        keywords=[ast.keyword(arg="metaclass", value=ast.Name(id=meta_name))],
        decorator_list=[],
        starargs=None,
        kwargs=None,
    )



TEXT_SCHEMA_CODE = '''\
import string
from typing import Any, cast, TypeAlias

from d42.custom_type import CustomSchema, PathHolder, Props, ValidationResult, register_type
from d42.custom_type.utils import make_substitution_error
from d42.custom_type.visitors import Generator, Representor, Substitutor, Validator
from district42.types import StrSchema
from district42.utils import TypeOrEllipsis
from niltype import Nil, Nilable


class TextProps(Props):
    @property
    def type(self) -> StrSchema:
        return cast(StrSchema, self.get("type", StrSchema()))


class _TextSchema(CustomSchema[TextProps]):
    type: TypeAlias = StrSchema
    name = "text"
    alphabet = string.ascii_letters
    specials = " " * 5

    def __call__(self, /, value: str) -> "_TextSchema":
        str_type = self.props.type(value)
        return self.__class__(self.props.update(type=str_type))

    def len(self, /, val_or_min: TypeOrEllipsis[int],
            max: Nilable[TypeOrEllipsis[int]] = Nil) -> "_TextSchema":
        str_type = self.props.type.len(val_or_min, max)
        return self.__class__(self.props.update(type=str_type))

    def __represent__(self, visitor: Representor, indent: int = 0) -> str:
        r = f"{visitor.name}.{self.name}"
        str_type = self.props.type

        if str_type.props.value is not Nil:
            r += f"({str_type.props.value!r})"

        if str_type.props.len is not Nil:
            r += f".len({str_type.props.len!r})"
        elif (str_type.props.min_len is not Nil) and (str_type.props.max_len is not Nil):
            r += f".len({str_type.props.min_len!r}, {str_type.props.max_len!r})"
        elif str_type.props.min_len is not Nil:
            r += f".len({str_type.props.min_len!r}, ...)"
        elif str_type.props.max_len is not Nil:
            r += f".len(..., {str_type.props.max_len!r})"

        return r

    def __generate__(self, visitor: Generator) -> str:
        str_type = self.props.type

        if str_type.props.value is not Nil:
            return str_type.props.value

        if str_type.props.len is not Nil:
            length = str_type.props.len
        else:
            min_length = str_type.props.min_len if (
                str_type.props.min_len is not Nil) else 0
            max_length = str_type.props.max_len if (
                str_type.props.max_len is not Nil) else 64
            length = visitor.random.random_int(min_length, max_length)

        generated = ""
        for idx in range(length):
            alphabet = self.alphabet + self.specials
            if (idx == 0) or (idx == length - 1):
                alphabet = self.alphabet
            elif len(generated) > 0 and (generated[-1] in self.specials):
                alphabet = self.alphabet
            generated += visitor.random.random_choice(alphabet)
        return generated

    def __validate__(self, visitor: Validator, value: Any, path: PathHolder) -> ValidationResult:
        return visitor.visit_str(self.props.type, value=value, path=path)

    def __substitute__(self, visitor: Substitutor, value: Any) -> "_TextSchema":
        result = visitor.validator.visit(self, value=value)
        if result.has_errors():
            raise make_substitution_error(result, visitor.formatter)
        str_type = self.props.type % value
        return self.__class__(self.props.update(type=str_type))


TextSchema = register_type(_TextSchema.name, _TextSchema)
'''



import ast
from typing import Dict, Tuple

from niltype import Nil

import ast_generate
from app.helpers import (
    get_module_to_import_from,
    get_types_from_any,
    is_builtin_class_instance,
    is_dict_typed_as_empty,
)
from app.modules.module import Import
from app.types._type import OverloadedFake, Typing, UnknownTypeSchema
from ast_generate import annotated_assign
from ast_generate.dict_value_type import (
    DictValueSubDictSimpleType,
    DictValueSubDictType,
    DictValueUnionType,
    ScalarDictValueType,
)


class DictTyping(Typing):

    @classmethod
    def is_valid_type_for_schema(cls, schema_: UnknownTypeSchema) -> bool:
        return schema_.class_name == 'DictSchema'

    def update_name(self, name: str):
        self.name = name

    def generate_pyi_empty_dict(self) -> Tuple[list[ast.AnnAssign], list[Import]]:
        annotation = annotated_assign(self.name, type(self.value).__name__)
        import_ = Import(get_module_to_import_from(self.value), self.value.__class__.__name__)
        return [annotation], [import_]

    def generate_pyi(self) -> Tuple[list[ast.expr | ast.AnnAssign], list[Import]]:
        imports = []
        annotations = []

        # todo для nested dict типизация присутствует дважды

        if is_dict_typed_as_empty(self.value):
            return self.generate_pyi_empty_dict()

        typing_map = {}

        for item in ['overload', 'Literal', 'TypedDict']:
            imports.append(Import('typing', item))

        for key, item_value in self.value.props.keys.items():
            value_, _ = item_value
            value_as_class = UnknownTypeSchema(key, value_)

            # если значение словаря само является словарем
            if DictTyping.is_valid_type_for_schema(value_as_class):
                sub_dict_schema = DictTyping(key, value_)

                if is_dict_typed_as_empty(sub_dict_schema.value):
                    imports.append(Import('typing', 'Dict'))
                    typing_map[key] = DictValueSubDictSimpleType(Dict)

                else:
                    sub_dict_schema.update_name(f'{self.name}_{key.capitalize()}Schema')

                    nested_annotations, nested_imports = sub_dict_schema.generate_pyi()
                    annotations.extend(nested_annotations)
                    imports.extend(nested_imports)

                    typing_map[key] = DictValueSubDictType(sub_dict_schema.name)

            elif (
                    value_.type is not None
                    and not is_builtin_class_instance(value_.type)
            ):
                value_type = value_.type

                # когда value = schema.any(...)
                if value_.__class__.__name__ == 'AnySchema':

                    if value_.props.types is not Nil:
                        types_in_any = get_types_from_any(value_.props)

                        if len(types_in_any) == 1:
                            type_ = types_in_any.pop()
                            imports.append(Import(get_module_to_import_from(type_), type_.__name__))
                            typing_map[key] = ScalarDictValueType(type_)

                        else:
                            imports.append(Import('typing', 'Union'))
                            for type_ in types_in_any:
                                imports.append(
                                    Import(get_module_to_import_from(type_), type_.__name__))
                            typing_map[key] = DictValueUnionType(types_in_any)

                    else:
                        # когда value = schema.any
                        imports.append(Import(get_module_to_import_from(value_), value_.__class__.__name__))
                        typing_map[key] = ScalarDictValueType(value_.__class__)

                else:
                    # когда value = typing.Any
                    imports.append(
                        Import(get_module_to_import_from(value_.type), value_type.__name__))
                    typing_map[key] = DictValueSubDictSimpleType(value_type)

            else:
                imports.append(
                    Import(get_module_to_import_from(value_), value_.__class__.__name__))
                typing_map[key] = ScalarDictValueType(value_.__class__)

        meta_class_name = f'_D42Meta{self.name}'
        annotations.append(ast_generate.dict_metaclass(meta_class_name, typing_map))
        annotations.append(ast_generate.dict_typeclass(self.name, meta_class_name, typing_map))

        return annotations, imports

    def generate_fake_overload(self, path_to_schema: str) -> Tuple[OverloadedFake, list[Import]]:

        if is_dict_typed_as_empty(self.value):
            imports = [
                Import('typing', 'Dict'),
                Import(get_module_to_import_from(self.value), self.class_name)
            ]
            overload = OverloadedFake(
                self.class_name,
                ast_generate.fake_scalar_overload(self.class_name, self.class_type)
            )
            return overload, imports

        module_name = path_to_schema.replace('/', '.').replace('.py', '')
        imports = [
            Import('typing', 'Type'),
            Import(module_name, self.name),
        ]
        overload = OverloadedFake(
            self.name,
            ast_generate.fake_dict_overload(self.name)
        )
        return overload, imports



import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'TestSchema'

CODE = '''\
from d42 import schema

TestSchema = schema.dict({
    "id": schema.int,
    "name[first]": schema.str,
})
'''

CODE_PYI = '''\
from district42.types import DictSchema
TestSchema: DictSchema\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from typing import Dict
from district42.types import DictSchema

@overload
def fake(schema: DictSchema) -> Dict:
    pass\
'''


def test_dict_with_breakets_pyi():
    module = load_module_from_string('test', CODE)
    schema_value = getattr(module, SCHEMA_NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(SCHEMA_NAME, schema_value)

    assert typed_module.get_printable_content() == CODE_PYI


def test_dict_with_breakets_pyi_blahblah():
    module = load_module_from_string('test.module', CODE)
    schema_value = getattr(module, SCHEMA_NAME)

    blahblha_module = modules.BlahBlahModule()
    blahblha_module.generate('test.module', SCHEMA_NAME, schema_value)

    assert blahblha_module.get_printable_content() == CODE_BLAHBLAH_PYI



import ast


class AnyDictValueType:

    def get_type_for_typeclass(self):
        pass

    def get_type_for_metaclass(self):
        pass


class ScalarDictValueType(AnyDictValueType):
    def __init__(self, type_):
        self.type_ = type_

    def get_type_for_typeclass(self) -> ast.Attribute:
        """ : IntSchema.type"""
        return ast.Attribute(
            value=ast.Name(id=self.type_.__name__),
            attr='type',
        )

    def get_type_for_metaclass(self) -> ast.Name:
        """ -> IntSchema"""
        return ast.Name(id=self.type_.__name__)


class DictValueUnionType(AnyDictValueType):
    def __init__(self, types_in_union):
        self.types_in_union = types_in_union

    def get_type_for_typeclass(self) -> ast.Subscript:
        """: Union[IntSchema.type, NoneSchema.type] """
        return ast.Subscript(
            value=ast.Name(id='Union'),
            slice=ast.Tuple(
                elts=[
                    ast.Attribute(
                        value=ast.Name(id=schema_in_union.__name__),
                        attr='type'
                    )
                    for schema_in_union in self.types_in_union
                ]
            )
        )

    def get_type_for_metaclass(self) -> ast.Subscript:
        """-> Union[IntSchema, NoneSchema]"""
        return ast.Subscript(
            value=ast.Name(id='Union'),
            slice=ast.Tuple(
                elts=[
                    ast.Name(id=schema_in_union.__name__)
                    for schema_in_union in self.types_in_union
                ]
            )
        )


class DictValueSubDictType(AnyDictValueType):

    def __init__(self, name):
        self.name = name

    def get_type_for_metaclass(self) -> ast.Name:
        """-> NestedTestSchema_EntitySchema"""
        return ast.Name(id=self.name)

    def get_type_for_typeclass(self) -> ast.Attribute:
        """-> NestedTestSchema_EntitySchema.type"""
        return ast.Attribute(
            value=ast.Name(id=self.name),
            attr='type',
        )


class DictValueSubDictSimpleType(AnyDictValueType):

    def __init__(self, type_):
        self.type_ = type_

    def get_type_for_metaclass(self) -> ast.Name:
        """ -> Dict"""
        return ast.Name(id=self.type_.__name__)

    def get_type_for_typeclass(self) -> ast.Name:
        """ : Dict"""
        return ast.Name(id=self.type_.__name__)



from __future__ import annotations

from setuptools import find_packages, setup


def find_required():
    with open('requirements.txt') as f:
        return f.read().splitlines()


def find_dev_required():
    with open("requirements-dev.txt") as f:
        return f.read().splitlines()


setup(
    name="d42-typing",
    version="0.0.2",
    description=".pyi typing stubs generation for d42 schemas",
    url="https://github.com/mytestopia/d42-typing",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=find_required(),
    tests_require=find_dev_required(),
    entry_points={
        'console_scripts': ['d42-typing=app.main:main']
    },
    author="Anna",
    author_email="testopia13@gmail.com",
    packages=find_packages(exclude=["tests"]),
    python_requires='>=3.10',
    license="Apache-2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



import ast
from typing import Tuple

from niltype import Nil

import ast_generate
from app.helpers import get_module_to_import_from
from app.modules.module import Import
from app.types._type import OverloadedFake, Typing, UnknownTypeSchema


class ListTyping(Typing):

    @classmethod
    def is_valid_type_for_schema(cls, schema_: UnknownTypeSchema) -> bool:
        return schema_.class_name == 'ListSchema'

    def generate_pyi(self) -> Tuple[list[ast.AnnAssign], list[Import]]:

        if self.value.props.type is not Nil:
            item_type = self.value.props.type
            item_type_class_name = self.value.props.type.__class__.__name__

            imports = [
                Import('typing', 'List'),
                Import(get_module_to_import_from(item_type), item_type_class_name)
            ]
            annotation = ast_generate.list_typeclass(self.name, item_type_class_name)
            return [annotation], imports

        annotation = ast_generate.annotated_assign(self.name, type(self.value).__name__)
        import_ = Import(get_module_to_import_from(self.value), self.value.__class__.__name__)

        return [annotation], [import_]

    def generate_fake_overload(self,  path_to_schema: str) -> Tuple[OverloadedFake, list[Import]]:
        imports = [Import('typing', 'List')]

        if self.value.props.type is not Nil:
            list_item_type = self.value.props.type.type
            list_item_type_name = list_item_type.__name__

            module_name = path_to_schema.replace('/', '.').replace('.py', '')
            imports.extend([
                Import(module_name, self.name),
                Import('typing', 'Type')
            ])
            overload = OverloadedFake(
                self.class_name,
                ast_generate.fake_list_overload(self.name, list_item_type_name)
            )
            return overload, imports

        imports.append(Import(get_module_to_import_from(self.value), 'ListSchema'))
        overload = OverloadedFake(
            'ListSchema',
            ast_generate.fake_scalar_overload('ListSchema', self.value.__class__.type)
        )
        return overload, imports



import ast


def _fake_overload(annotation, returns):
    return ast.FunctionDef(
        name='fake',
        args=ast.arguments(
            posonlyargs=[],
            args=[ast.arg(
                arg='schema',
                annotation=annotation,
                type_comment=None,
            )],
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[]
        ),
        body=[ast.Pass()],
        decorator_list=[ast.Name(id='overload')],
        returns=returns
    )


def fake_none_overload(type_schema: str):
    """
    Возвращает метод перегрузки для скалярных типов:
    @overload
    def fake(schema: NoneSchema) -> None: pass
    """
    return _fake_overload(
        annotation=ast.Name(id=type_schema),
        returns=ast.Name(id='None')
    )


def fake_scalar_overload(input_type: str, output_type: str):
    """
    Возвращает метод перегрузки для скалярных типов:
    @overload
    def fake(schema: <input_type>) -> <output_type>: pass
    """
    return _fake_overload(
        annotation=ast.Name(id=input_type),
        returns=ast.Name(id=output_type.__name__),
    )


def fake_union_overload(type_schema: str, types_in_union):
    """
    Возвращает метод перегрузки для скалярных типов:
    @overload
    def fake(schema: type_schema) -> Union[...]: pass
    """
    def get_elt_2(type_: str | None) -> ast.Constant | ast.Name:
        if type_ is None:
            return ast.Constant(value=None)
        return ast.Name(id=type_.__name__)

    return _fake_overload(
        annotation=ast.Name(id=type_schema),
        returns=ast.Subscript(
            value=ast.Name(id='Union'),
            slice=ast.Tuple(
                elts=[
                    get_elt_2(type_.type) for type_ in types_in_union
                ],
            ),
        ),
    )


def fake_dict_overload(schema_name: str):
    """
    Возвращает метод перегрузки для сложных типов:
    @overload
    def fake(schema: Type[NameSchema]) -> NameSchema.type: pass
    """
    return _fake_overload(
        annotation=ast.Subscript(
            value=ast.Name(id='Type'),
            slice=ast.Name(id=schema_name),
        ),
        returns=ast.Attribute(value=ast.Name(id=schema_name), attr='type')
    )


def fake_list_overload(schema_name: str, item_type: str):
    """
    Возвращает метод перегрузки для сложных типов:
    @overload
    def fake(schema: Type[NameSchema]) -> List[str]
    """
    return _fake_overload(
        annotation=ast.Subscript(
            value=ast.Name(id='Type'),
            slice=ast.Name(id=schema_name),
        ),
        returns=ast.Subscript(
            value=ast.Name(id='List'),
            slice=ast.Name(id=item_type)
        )
    )



import ast
from typing import Any

from ast_generate.dict_value_type import AnyDictValueType


def get_getitem_method_return_ast(type_schema: Any) -> Any:
    """
    Возвращает ast-тип возвращаемого значения метода __getitem__.
    """
    return type_schema.get_type_for_metaclass()


def getitem_method(key_name: str, type_schema: Any, keys_count: int):
    """
    Возвращает ast-метод, используемый в мета-классе для типизации словаря:
    @overload
    def __getitem__(cls, arg: Literal["id"]) -> IntSchema:
        pass
    """
    decorators_list = [ast.Name(id='overload')] if keys_count > 1 else []
    return ast.FunctionDef(
        name='__getitem__',
        args=[
            ast.arguments(
                posonlyargs=[],
                args=[
                    ast.arg(arg='cls'),
                    ast.arg(
                        arg='arg',
                        annotation=ast.Subscript(
                            value=ast.Name(id='Literal'),
                            slice=ast.Constant(value=key_name, kind=None),
                        ),
                        type_comment=None,
                    ),
                ],
                defaults=[],
                kwonlyargs=[]
            )
        ],
        body=[ast.Pass()],
        decorator_list=decorators_list,
        returns=get_getitem_method_return_ast(type_schema)
    )


def schema_substitution_methods() -> list[ast.FunctionDef]:
    """
    Возвращает список ast-методов, используемый в мета-классе для типизации словаря:
    def __mod__(self, other): pass
    def __add__(self, other): pass
    """
    def ast_method(name: str) -> ast.FunctionDef:
        return ast.FunctionDef(
            name=name,
            args=[
                ast.arguments(
                    posonlyargs=[],
                    args=[ast.arg(arg='self'), ast.arg(arg='other')],
                    defaults=[],
                    kwonlyargs=[]
                )
            ],
            body=[ast.Pass()],
            decorator_list=[],
        )
    return [ast_method(name) for name in ['__mod__', '__add__']]


def dict_metaclass(name: str, typing_map: dict[str, AnyDictValueType]):
    """
    Возвращает ast-класс, используемый в мета-классе для типизации словаря.
    Пример:
    class _D42MetaObjectSchema(type):

        @overload
        def __getitem__(cls, arg: Literal['id']) -> int:
            pass

        @overload
        def __getitem__(cls, arg: Literal['name']) -> str:
            pass

        def __mod__(self, other):
            pass

        def __add__(self, other):
            pass
    """
    return ast.ClassDef(
        name=f"{name}",  # gen_unique_name
        bases=[ast.Name(id="type")],
        body=[
                 getitem_method(key, value, len(typing_map.keys()))
                 for key, value in typing_map.items()
             ] + schema_substitution_methods(),
        keywords=[],
        decorator_list=[]
    )



from app.types._type import Typing
from ast_generate import annotated_assign


class ConstantTyping(Typing):

    def generate_pyi(self):
        return annotated_assign(self.name, type(self.value).__name__)



import sys

import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'NestedTestSchema'

CODE_ENTITY = '''\
from d42 import schema

EntitySchema = schema.dict({
    'is_deleted': schema.bool,
    'updated_at': schema.int
})
'''

CODE = '''\
from d42 import schema
from test_entity import EntitySchema

NestedTestSchema = schema.dict({
    'id': schema.int,
    'entity': EntitySchema
})
'''

CODE_PYI = '''\
from typing import overload
from typing import Literal
from typing import TypedDict
from district42.types import IntSchema
from district42.types import BoolSchema

class _D42MetaNestedTestSchema_EntitySchema(type):

    @overload
    def __getitem__(cls, arg: Literal['is_deleted']) -> BoolSchema:
        pass

    @overload
    def __getitem__(cls, arg: Literal['updated_at']) -> IntSchema:
        pass

    def __mod__(self, other):
        pass

    def __add__(self, other):
        pass

class NestedTestSchema_EntitySchema(metaclass=_D42MetaNestedTestSchema_EntitySchema):

    class type(TypedDict, total=False):
        is_deleted: BoolSchema.type
        updated_at: IntSchema.type

class _D42MetaNestedTestSchema(type):

    @overload
    def __getitem__(cls, arg: Literal['id']) -> IntSchema:
        pass

    @overload
    def __getitem__(cls, arg: Literal['entity']) -> NestedTestSchema_EntitySchema:
        pass

    def __mod__(self, other):
        pass

    def __add__(self, other):
        pass

class NestedTestSchema(metaclass=_D42MetaNestedTestSchema):

    class type(TypedDict, total=False):
        id: IntSchema.type
        entity: NestedTestSchema_EntitySchema.type\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from typing import Type
from test.module import NestedTestSchema

@overload
def fake(schema: Type[NestedTestSchema]) -> NestedTestSchema.type:
    pass\
'''


class TestClass:
    def setup_method(self):
        entity_module = load_module_from_string('test_entity', CODE_ENTITY)
        sys.modules['test_entity'] = entity_module

    def teardown_method(self):
        sys.modules.pop('test_entity')

    def test_dict_nested_imported_pyi(self):
        module = load_module_from_string('test_scalar', CODE)
        schema_value = getattr(module, SCHEMA_NAME)

        typed_module = modules.TypedModule('file_name')
        typed_module.generate(SCHEMA_NAME, schema_value)

        assert typed_module.get_printable_content() == CODE_PYI

    def test_dict_nested_imported_pyi_blahblah(self):
        module = load_module_from_string('test.module', CODE)

        schema_value = getattr(module, SCHEMA_NAME)

        blahblah_module = modules.BlahBlahModule()
        blahblah_module.generate('test.module', SCHEMA_NAME, schema_value)

        assert blahblah_module.get_printable_content() == CODE_BLAHBLAH_PYI



import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'TestSchema'

CODE = '''\
from d42 import schema
TestSchema = schema.list(schema.int)
'''

CODE_PYI = '''\
from typing import List
from district42.types import IntSchema

class TestSchema:
    type = List[IntSchema.type]\
'''

BLAHBLAH_PYI = '''\
from typing import overload
from typing import List
from test.module import TestSchema
from typing import Type

@overload
def fake(schema: Type[TestSchema]) -> List[int]:
    pass\
'''


def test_list_of_int_pyi():
    module = load_module_from_string('test_scalar', CODE)
    schema_description = getattr(module, SCHEMA_NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(SCHEMA_NAME, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_list_of_int_pyi_blahblah():
    module = load_module_from_string('test_scalar', CODE)
    schema_description = getattr(module, SCHEMA_NAME)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test.module', SCHEMA_NAME, schema_description)

    assert blahblah_module.get_printable_content() == BLAHBLAH_PYI



import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'TestSchema'

CODE = '''\
from d42 import schema
TestSchema = schema.any
'''

CODE_PYI = '''\
from district42.types import AnySchema
TestSchema: AnySchema\
'''

BLAHBLAH_PYI = '''\
from typing import overload
from typing import Any
from district42.types import AnySchema

@overload
def fake(schema: AnySchema) -> Any:
    pass\
'''


def test_any_pyi():
    module = load_module_from_string('test_scalar', CODE)
    schema_description = getattr(module, SCHEMA_NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(SCHEMA_NAME, schema_description)

    assert typed_module.get_printable_content() == CODE_PYI


def test_any_pyi_blahblah_any():
    module = load_module_from_string('test_scalar', CODE)
    schema_description = getattr(module, SCHEMA_NAME)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test_file_name', SCHEMA_NAME, schema_description)

    assert blahblah_module.get_printable_content() == BLAHBLAH_PYI



import typing

from d42.custom_type import CustomSchema
from district42.types import AnyProps


def is_builtin_class_instance(obj):
    return obj.__module__ == 'builtins'


def is_schema_type_simple(schema_value: typing.Any) -> bool:
    """Т.е. тип схемы не является пользовательским классом."""
    # для NoneSchema нельзя метод is_builtin_class_instance падает с исключением
    return (
            schema_value.type is None
            or schema_value.type is typing.Any
            or schema_value.type is typing.Dict
            or schema_value.type is typing.List
            or is_builtin_class_instance(schema_value.type)
    )


def get_module_to_import_from(schema_value: typing.Any) -> str:

    def remove_protected_prefix(module_name: str) -> str:
        """district42.types._any_schema -> district42.types"""
        parts = module_name.split('.')
        filtered_parts = [part for part in parts if not part.startswith('_')]
        return '.'.join(filtered_parts)

    import_from = schema_value.__module__
    return remove_protected_prefix(import_from)


def get_types_from_any(any_value_props) -> list[typing.Any]:
    types_list = list()
    for type_ in any_value_props.types:
        if issubclass(type_.__class__, CustomSchema):
            if (
                hasattr(type_, 'type') is False
                or type_.type is typing.Any
            ):
                types_list.append(typing.Any)
            else:
                types_list.append(type_.type)

        elif isinstance(type_.props, AnyProps):
            types_list.extend(get_types_from_any(type_.props))

        else:
            types_list.append(type_.__class__)

    unique_types = []
    [unique_types.append(x) for x in types_list if x not in unique_types]
    return unique_types



import app.modules as modules
from app.helpers import load_module_from_string

SCHEMA_NAME = 'NestedTestSchema'

CODE = '''\
from d42 import schema

NestedTestSchema = schema.dict({
    'id': schema.int,
    'entity': schema.dict({
        'is_deleted': schema.bool,
        'updated_at': schema.int
    })
})
'''

CODE_PYI = '''\
from typing import overload
from typing import Literal
from typing import TypedDict
from district42.types import IntSchema
from district42.types import BoolSchema

class _D42MetaNestedTestSchema_EntitySchema(type):

    @overload
    def __getitem__(cls, arg: Literal['is_deleted']) -> BoolSchema:
        pass

    @overload
    def __getitem__(cls, arg: Literal['updated_at']) -> IntSchema:
        pass

    def __mod__(self, other):
        pass

    def __add__(self, other):
        pass

class NestedTestSchema_EntitySchema(metaclass=_D42MetaNestedTestSchema_EntitySchema):

    class type(TypedDict, total=False):
        is_deleted: BoolSchema.type
        updated_at: IntSchema.type

class _D42MetaNestedTestSchema(type):

    @overload
    def __getitem__(cls, arg: Literal['id']) -> IntSchema:
        pass

    @overload
    def __getitem__(cls, arg: Literal['entity']) -> NestedTestSchema_EntitySchema:
        pass

    def __mod__(self, other):
        pass

    def __add__(self, other):
        pass

class NestedTestSchema(metaclass=_D42MetaNestedTestSchema):

    class type(TypedDict, total=False):
        id: IntSchema.type
        entity: NestedTestSchema_EntitySchema.type\
'''

CODE_BLAHBLAH_PYI = '''\
from typing import overload
from typing import Type
from test.module import NestedTestSchema

@overload
def fake(schema: Type[NestedTestSchema]) -> NestedTestSchema.type:
    pass\
'''


def test_dict_nested_pyi():
    module = load_module_from_string('test_scalar', CODE)

    schema_value = getattr(module, SCHEMA_NAME)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(SCHEMA_NAME, schema_value)

    assert typed_module.get_printable_content() == CODE_PYI


def test_dict_nested_pyi_blahblah():
    module = load_module_from_string('test.module', CODE)

    schema_value = getattr(module, SCHEMA_NAME)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test.module', SCHEMA_NAME, schema_value)

    assert blahblah_module.get_printable_content() == CODE_BLAHBLAH_PYI



import pytest

import app.modules as modules
from app.helpers import load_module_from_string


@pytest.mark.parametrize('schema, schema_type', [
    ('schema.str', 'StrSchema'),
    ('schema.float', 'FloatSchema'),
    ('schema.bool', 'BoolSchema'),
    ('schema.int', 'IntSchema'),
    ('schema.bytes', 'BytesSchema'),
    ('schema.none', 'NoneSchema'),
    ('schema.datetime', 'DateTimeSchema'),
])
def test_scalar_pyi(schema, schema_type):
    code = """from d42 import schema\n\n\nTestSchema = {}""".format(schema)
    module = load_module_from_string('test_scalar', code)

    schema_name = 'TestSchema'
    schema_description = getattr(module, schema_name)

    typed_module = modules.TypedModule('file_name')
    typed_module.generate(schema_name, schema_description)

    assert typed_module.get_printable_content() == (
        f'from district42.types import {schema_type}\n'
        f'TestSchema: {schema_type}'
    )


@pytest.mark.parametrize('schema, input_type, output_type', [
    ('schema.str', 'StrSchema', 'str'),
    ('schema.float', 'FloatSchema', 'float'),
    ('schema.bool', 'BoolSchema', 'bool'),
    ('schema.int', 'IntSchema', 'int'),
    ('schema.bytes', 'BytesSchema', 'bytes'),
    ('schema.none', 'NoneSchema', 'None'),
    ('schema.datetime', 'DateTimeSchema', 'str'),
])
def test_scalar_pyi_blahblah(schema, input_type, output_type):
    code = """from d42 import schema\n\n\nTestSchema = {}""".format(schema)
    module = load_module_from_string('test_scalar', code)

    schema_name = 'TestSchema'
    schema_description = getattr(module, schema_name)

    blahblah_module = modules.BlahBlahModule()
    blahblah_module.generate('test_file_name', schema_name, schema_description)

    assert blahblah_module.get_printable_content() == (
        f'from typing import overload\n'
        f'from district42.types import {input_type}\n'
        f'\n'
        f'@overload\n'
        f'def fake(schema: {input_type}) -> {output_type}:\n'
        f'    pass'
    )
